#ifdef __USAGE
%C - Display/Write PCI Latency/Command Registers.

%C	[options*]

options:
 -d device id   Device Id of adapter.
 -v vendor id   Vendor Id of adapter.
 -p pci index   PCI index (0-number of devices).
 -c command     Value to write to command register.
 -l latency     Value to write to latency register.
 -V             Verbose.

#endif

#include <stdio.h>
#include <stdlib.h>
#include <i86.h>
#include <string.h>
#include <stddef.h>
#include <unistd.h>
#include <sys/pci.h>

#define		OPTION_VALID	0x80000000

main( int argc, char *argv[] )
{
	unsigned		did, vid, busnum, devfuncnum, pindex = 0;
	unsigned		num_buses, version, hardware;
	long			command = 0, latency = 0, temp;
	int				option, verbose = 0;

	while( ( option = getopt(argc, argv, "Vv:d:l:c:p:" ) ) != EOF ) {
		switch( option ) {
		    case 'V':
				verbose++;
				break;
		    case 'v':
				vid = atoh(optarg);
				break;
			case 'd':
				did = atoh(optarg);
				break;
		    case 'l':
				latency = atoh(optarg) | OPTION_VALID;
				break;
			case 'c':
				command = atoh(optarg) | OPTION_VALID;
				break;
			case 'p':
				pindex = atoh(optarg);
				break;
			default:
				break;
		}
	}
	
	if( _CA_PCI_BIOS_Present( &num_buses, &version, &hardware ) != PCI_SUCCESS ) {
		printf("PCI BIOS not found\n");
		exit( 0 );
	}

	if( _CA_PCI_Find_Device( did, vid, pindex, &busnum, &devfuncnum) != PCI_SUCCESS ) {
		printf("Unable to locate device\n");
		exit( 0 );
	}

	temp = 0;
	if( _CA_PCI_Read_Config_Word( busnum, devfuncnum,
					offsetof(struct _pci_config_regs, Command),
					1, (char *)&temp) != PCI_SUCCESS ) {
		printf("Unable to read command register\n");
		exit( 0 );
	}

	printf("Command Register Current Value: %x\n", temp);

	if( verbose ) {
		command_reg_info( temp );
	}

	if( command & OPTION_VALID ) {
		if( _CA_PCI_Write_Config_Word( busnum, devfuncnum,
						offsetof(struct _pci_config_regs, Command),
						1, (char *)&command ) != PCI_SUCCESS) {
			printf("Unable to write command register\n");
			exit( 0 );
		}
		printf("Command Register New Value: %x\n", command & 0xffff );
	}

	temp = 0;
	if( _CA_PCI_Read_Config_Byte( busnum, devfuncnum, offsetof(struct _pci_config_regs, Latency_Timer),
									1, (char *)&temp ) != PCI_SUCCESS ) {
		printf("Unable to read latency register\n");
		exit( 0 );
	}

	printf("\nLatency Register Current Value: %x\n", temp);

	if( latency & OPTION_VALID ) {
		if( _CA_PCI_Write_Config_Byte( busnum, devfuncnum, offsetof(struct _pci_config_regs, Latency_Timer),
											1, (char *)&latency ) != PCI_SUCCESS) {
			printf("Unable to write latency register\n");
			exit( 0 );
		}
		printf("Latency Register New Value: %x\n", latency & 0xffff );
	}
}

command_reg_info( long temp )
{
	printf("	I/O space access %s\n", (temp & 1) ? "enabled" : "disabled" );
	printf("	Memory space access %s\n", (temp & 2) ? "enabled" : "disabled" );
	printf("	Bus Master %s\n", (temp & 4) ? "enabled" : "disabled" );
	printf("	Special Cycle opertations %s\n", (temp & 8) ? "monitor" : "ignored" );
	printf("	Memory Write and Invalidate %s\n", (temp & 0x10) ? "enabled" : "disabled" );
	printf("	Palette Snooping %s\n", (temp & 0x20) ? "enabled" : "disabled" );
	printf("	Parity Checking %s\n", (temp & 0x20) ? "enabled" : "disabled" );
	printf("	Data/Address stepping %s\n", (temp & 0x40) ? "enabled" : "disabled" );
	printf("	SERR# driver %s\n", (temp & 0x80) ? "enabled" : "disabled" );
	printf("	Fast back-to-back transactions to different agents %s\n", (temp & 0x100) ? "enabled" : "disabled" );
}