#ifdef __USAGE
%C - Display PCI devices.

%C	[-v]

Where:
 -v             Verbose

#endif

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <i86.h>
#include <string.h>
#include <sys/seginfo.h>
#include <sys/osinfo.h>
#include <sys/pci.h>
#include <sys/inline.h>
int		verbose = 0;

unsigned long pci_io_mem_length(unsigned busnum, unsigned devfuncnum, unsigned pos, unsigned type);

char *classes[] = {
	"Pre-2.0", "Mass Storage", "Network", "Display", "Multimedia", "Memory", "Bridge",
};
char *pre[] = {
	"Non-VGA", "VGA",
};
char *mass[] = {
	"SCSI", "IDE", "Floppy", "IPI",
};
char *net[] = {
	"Ethernet", "Token ring", "FDDI",
};
char *disp[] = {
	"VGA", "XGA",
};
char *multi[] = {
	"Video", "Audio",
};
char *mem[] = {
	"RAM", "Flash",
};
char *bridge[] = {
	"Host/PCI", "PCI/ISA", "PCI/EISA", "PCI/MCA", "PCI/PCI", "PCI/PCMCIA",
};
char **class_table[] = {
	pre, mass, net, disp, multi, mem, bridge,
};

void dsp_vid_dev(unsigned long vendor, unsigned long device)
{
	unsigned long val;
	
	val = (device << 16) | vendor;

	printf("Vendor         = ");
	switch (vendor)
	{
		case 0x1002:
			printf ("ATI TECHNOLOGIES INC ");
			break;
		case 0x1000:
			printf ("NCR Corporation ");
			break;
		case 0x1045:
			printf ("OPTI ");
			break;
		case 0x8086:
			printf ("INTEL CORPORATION ");
			break;
		case 0x9004:
			printf ("Adaptec Inc. ");
			break;
		case 0x1022:
			printf ("Advanced Micro Devices ");
			break;
		case 0x102b:
			printf ("MATROX ");
			break;
		case 0x1011:
			printf ("Digital ");
			break;
	};

	printf("ID=%xh\n", vendor);

	printf("Device         = ");
	switch (val)
	{
		case 0x122d8086:
			printf ("82438FX PCI-host and dram controller (Triton) ");
			break;
		case 0x122e8086:
			printf ("82371FB PCI-ISA bridge (Triton) ");
			break;
		case 0x12308086:
			printf ("82371FB PCI fast-IDE (Triton) ");
			break;
		case 0x04848086:
			printf ("82378IB PCI-ISA bridge ");
			break;
		case 0x04838086:
			printf ("82424ZX cache dram controller ");
			break;
		case 0x04828086:
			printf ("82375EB PCI-EISA bridge ");
			break;
		case 0x04A38086:
			printf ("82434LX PCI cache memory controller ");
			break;
		case 0x50789004:
			printf ("AIC-7850 PCI SCSI host adapter ");
			break;
		case 0x71789004:
			printf ("AHA-2940 PCI SCSI host adapter ");
			break;
		case 0x72789004:
			printf ("AHA-3940 PCI SCSI host adapter ");
			break;
		case 0x00011000:
			printf ("NCR-53c810 PCI SCSI host adapter ");
			break;
		case 0x20001022:
			printf ("Am79c970 PCI Ethernet adapter ");
			break;
		case 0x20201022:
			printf ("Am-53c974 PCI SCSI host adapter ");
			break;
		case 0x0518102b:
			printf ("MGA PCI ");
			break;
		case 0x47581002:
			printf ("Mach 64 ");
			break;
		case 0x000f1011:
			printf ("FDDI controller ");
			break;
	}
	printf("ID=%xh\n", device);
}

void dspclass(char *class)
{
	unsigned 				i;
	char					**table;

	i = class[2];
	if(i > sizeof classes / sizeof *classes)
	{
		printf("Unknown(%xh)", class);
	}
	else
	{
		printf("%s ", classes[i]);
		table = class_table[i];
		i = class[1];
		while(i-- && *table)
		{
			table++;
		}
		if(*table)
		{
			printf("(%s)", *table);
		}
		else
		{
			printf("(Other %d)", class[1]);
		}
		if(verbose)
			printf(" ProgIF=%d", class[0]);
	}
}		

int showdevice_verbose(unsigned busnum, unsigned devfuncnum)
{
	int							i;
	struct _pci_config_regs		creg;
	const char *pin_names[5] = { "NC", "INT A", "INT B", "INT C", "INT D" };

	_CA_PCI_Read_Config_DWord(busnum, devfuncnum, 0, 16, (char *)&creg);

	if( creg.Vendor_ID == 0xffff )
		return( 0 );
	dsp_vid_dev(creg.Vendor_ID, creg.Device_ID);
	pci_index( creg.Device_ID, creg.Vendor_ID, busnum, devfuncnum );
	printf("Class Code     = %xh ", *(unsigned *)creg.Class_Code);
	dspclass(creg.Class_Code);
	printf("\n");
	printf("Revision ID    = %xh\n", creg.Revision_ID);
	printf("Bus number     = %d\n", busnum);
	printf("Device number  = %d\n", devfuncnum >> 3);
	printf("Function num   = %d\n", devfuncnum & 0x3);
	printf("Status Reg     = %xh\n", creg.Status);
	printf("Command Reg    = %xh\n", creg.Command);
	printf("Header type    = %xh %s-function\n", creg.Header_Type & 0x7f, (creg.Header_Type & 0x80) ? "Multi" : "Single");
	printf("BIST           = %xh ", creg.BIST);
	if(creg.BIST & 0x80)
	{
		if(creg.BIST & 0xf)
			printf("Selftest error %xh\n", creg.BIST & 0x0f);
		else
			printf("Selftest OK\n");
	}
	else
		printf("Build-in-self-test not supported\n");
	printf("Latency Timer  = %xh\n", creg.Latency_Timer);
	printf("Cache Line Size= %xh %s\n", creg.Cache_Line_Size, creg.Cache_Line_Size ? "un-cacheable" : "");
	switch(creg.Header_Type & 0x7f)
	{
		case 0:
			for(i = 0; i < 6; i++)
			{
				if(creg.Base_Address_Regs[i] == 0)
					break;
				if(i==0 && creg.Base_Address_Regs[i])
					printf("Base Address   =");
				if(PCI_IS_IO(creg.Base_Address_Regs[i]))
				{
					printf(" IO@%xh", PCI_IO_ADDR(creg.Base_Address_Regs[i]));
					if(creg.Base_Address_Regs[i] & 0x02)
						printf(",reserved");
					printf(" length %ld bytes", pci_io_mem_length(busnum, devfuncnum, 0x10+(i*4), 1));
				}
				else
				{
					printf(" MEM@%xh", PCI_MEM_ADDR(creg.Base_Address_Regs[i]));
					if(creg.Base_Address_Regs[i] & 0x08)
						printf(",Prefetchable");

					if(PCI_IS_MMAP20(creg.Base_Address_Regs[i]))
						printf(",20bit");
					else if(PCI_IS_MMAP32(creg.Base_Address_Regs[i]))
						printf(",32bit");
					else if(PCI_IS_MMAP64(creg.Base_Address_Regs[i]))
						printf(",64bit");
					else
						printf(",reserved");
					printf(" length %ld", pci_io_mem_length(busnum, devfuncnum, 0x10+(i*4), 1));
				}
			}
			if(i)
				printf("\n");
			if(creg.ROM_Base_Address)
			{
				printf("Expansion ROM  = %xh ", PCI_ROM_ADDR(creg.ROM_Base_Address));
				printf(" length %ld", pci_io_mem_length(busnum, devfuncnum, 0x30, 0));
				if((creg.ROM_Base_Address & 0x1) || (creg.Command & 0x2))
					printf(" enabled\n");
				else
					printf(" disabled\n");
			}
			printf("Max Lat        = %dns\n", creg.Max_Lat);
			printf("Min Gnt        = %dns\n", creg.Min_Gnt);
			printf("PCI Int Pin    = %d\n", creg.Interrupt_Pin);
			printf("PCI Int Pin    = %s\n", pin_names[creg.Interrupt_Pin]);
			if(creg.Interrupt_Line <= 0xf)
				printf("Interrupt line = %d\n", creg.Interrupt_Line);
			else if(creg.Interrupt_Line == 0xff)
				printf("Interrupt line = no connection\n");
			else
				printf("Interrupt line = reserved\n");
			break;

		default:
			printf("Unknown header type\n");
			break;
	}
	printf("\n");
	return 0;
}

int showdevice(unsigned busnum, unsigned devfuncnum)
{
	int							i, flag;
	struct _pci_config_regs		creg;
	const char *pin_names[5] = { "NC", "INT A", "INT B", "INT C", "INT D" };

	_CA_PCI_Read_Config_DWord(busnum, devfuncnum, 0, 16, (char *)&creg);

	if( creg.Vendor_ID == 0xffff )
		return( 0 );

	if(creg.Class_Code[2] < 1 || creg.Class_Code[2] > 4)
		return 0;

	printf("Class          = ");
	dspclass(creg.Class_Code);
	printf("\n");
	dsp_vid_dev(creg.Vendor_ID, creg.Device_ID);
	pci_index( creg.Device_ID, creg.Vendor_ID, busnum, devfuncnum );
	flag = 0;
	for(i = 0; i < 6; i++)
	{
		if(creg.Base_Address_Regs[i]==0)
			break;
		if(!flag) {
			printf("                 ");
			flag = 1;
		}
		if(PCI_IS_IO(creg.Base_Address_Regs[i]) & 1)
			printf("IO @ %xh ", PCI_IO_ADDR(creg.Base_Address_Regs[i]));
		else
			printf("MEM @ %xh ", PCI_MEM_ADDR(creg.Base_Address_Regs[i]));
	}
	if(creg.ROM_Base_Address)
	{
		if(!flag) {
			printf("                 ");
		}
		printf("ROM @ %xh ", PCI_ROM_ADDR(creg.ROM_Base_Address));
		if((creg.ROM_Base_Address & 0x1) || (creg.Command & 0x2))
			printf("enabled\n");
		else
			printf("disabled\n");
	}
	else
	{
		if(i)
			printf("\n");
	}
	printf("PCI Int Pin    = %s\n", pin_names[creg.Interrupt_Pin]);
	if(creg.Interrupt_Line <= 0xf)
		printf("Interrupt line = %d\n", creg.Interrupt_Line);
	else if(creg.Interrupt_Line == 0xff)
		printf("Interrupt line = no connection\n");
	else
		printf("Interrupt line = reserved\n");
	printf("\n");
	return 0;
}

unsigned long pci_io_mem_length(unsigned busnum, unsigned devfuncnum, unsigned pos, unsigned type)
{
	unsigned long	cur_addr, size, tmp, shift;
	
	_CA_PCI_Read_Config_DWord(busnum, devfuncnum, pos, 1, (char *)&cur_addr);
	tmp = 0xffffffff;
	_CA_PCI_Write_Config_DWord(busnum, devfuncnum, pos, 1, (char *)&tmp);
	_CA_PCI_Read_Config_DWord(busnum, devfuncnum, pos, 1, (char *)&tmp);

	if(type)
	{
		if(tmp & 0x1)
			tmp &=0xfffffffc;
		else
			tmp &=0xfffffff0;
	}
	else
		tmp &=0xfffff800;

	for(shift=0;shift<32;shift++)
	{
		size = ((unsigned long)1)<<shift;
		if((tmp>>shift) & 0x1) 
			break;
	}
	
	_CA_PCI_Write_Config_DWord(busnum, devfuncnum, pos, 1, (char *)&cur_addr);
	return size;
}

main(int argc, char *argv[])
{
	unsigned lastbus, busnum, device, function, num_fcns;
	unsigned header_type, device_id, version, hardware;
	struct _osinfo info;

	argv = argv;

	if(argc > 1)
		verbose++;
	
	if(qnx_osinfo(0, &info) == -1)
	{
		printf("NO BIOS32 found\n");
		exit(0);
	}

	if(!(info.sflags & _PSF_PCI_BIOS))		// we have a PCI BIOS
	{
		printf("NO BIOS32 found\n");
		exit(0);
	}

	if(_CA_PCI_BIOS_Present(&lastbus, &version, &hardware) != PCI_SUCCESS)
	{
		printf("NO BIOS32 found\n");
		exit(0);
	}

	printf("\nPCI version    = %d.%02x\n\n", (version >> 8) & 0xff, version & 0xff);

	if(lastbus > 0x40)			// KLUDGE FOR AMI BIOS it reports 66 buses
		lastbus = 0;

	for(busnum = 0; busnum <= lastbus; busnum++)
	{
		for(device = 0; device < 32; device++)
		{
			_CA_PCI_Read_Config_Word(busnum, device << 3, offsetof(struct _pci_config_regs, Device_ID),
									1, (char *)&device_id);
			if(device_id == 0xffff)
				continue;
			_CA_PCI_Read_Config_Byte(busnum, device << 3, offsetof(struct _pci_config_regs, Header_Type),
									1, (char *)&header_type);
			if(header_type & 0x80) {
				num_fcns = 8;
			} else {
				num_fcns = 1;
			}
			for(function = 0; function < num_fcns; function++)
			{
				if(verbose)
					showdevice_verbose(busnum, (device << 3) | function);
				else
					showdevice(busnum, (device << 3) | function);
			}
		}
	}
	
	if(verbose)
	{
		showirq();
	}
}

struct _pci_irq_routing {
	unsigned char						bus;
	unsigned char						device;
	struct _pci_irq_info {
		unsigned char						link;
		unsigned short						mask;
	}									irq[4];
	unsigned char						slot;
	unsigned char						reserved;
};

void showirq(void) {
	struct _pci_route_buffer			cmd;
	unsigned							irq;
	struct _pci_irq_routing				*route, *r, *r2;
	int									i, i2, j, j2;

	memset(&cmd, 0x00, sizeof cmd);
	if(_CA_PCI_Get_Routing_Options(&cmd, 0) != PCI_BUFFER_TOO_SMALL) {
		return;
	}
	if(!(route = malloc(cmd.BufferSize))) {
		return;
	}
	cmd.DataBufferOffset = FP_OFF(route);
	cmd.DataBufferSeg = FP_SEG(route);
	
	if(_CA_PCI_Get_Routing_Options(&cmd, &irq) != PCI_SUCCESS) {
		return;
	}

	for(r = route, i = 0; i < cmd.BufferSize / sizeof *route; i++, r++) {
		struct _pci_config_regs		creg;
		int							num_fcns;

		printf("IRQ Routing    = bus=%d", r->bus);
		if(r->slot) {
			printf(" slot=%d", r->slot);
		} else {
			printf(" motherboard");
		}
		printf(" device=%d", r->device >> 3);
		if(r->device & 0x07) {
			printf(" func=%d", r->device & 0x07);
		}
		printf("\n");

		for(num_fcns = 1, j = 0; j < num_fcns; j++) {
			_CA_PCI_Read_Config_DWord(r->bus, (r->device & 0xf8) | j, 0, 16, (char *)&creg);
			if(creg.Vendor_ID != 0xffff) {
				if(creg.Header_Type & 0x80) {
					num_fcns = 8;
				}
				dsp_vid_dev(creg.Vendor_ID, creg.Device_ID);
			} else if(num_fcns == 1) {
				printf("                 No card configured in slot\n");
			}
		}

		for(j = 0; j < sizeof r->irq / sizeof *r->irq; j++) {
			int								flag;

			printf("                ");
			if(r->irq[j].link) {
				printf(" INT%c=", 'A' + j);
				if(r->irq[j].mask) {
					int							b;

					for(flag = b = 0; b < 16; b++) {
						if(r->irq[j].mask & (1 << b)) {
							if(flag) {
								printf(",");
							} else {
								flag = 1;
							}
							printf("%d", b);
						}
					}
				} else {
					printf("NC");
				}
				for(r2 = route, flag = i2 = 0; i2 < cmd.BufferSize / sizeof *route; i2++, r2++) {
					for(flag = j2 = 0; j2 < sizeof r->irq / sizeof *r->irq; j2++) {
						if(r->irq[j].link == r2->irq[j2].link && (i != i2 || j != j2)) {
							if(flag) {
								printf(",");
							} else {
								flag = 1;
								if(r2->slot) {
									printf(" Slot%d", r2->slot);
								} else {
									printf(" Motherboard");
								}
								printf(":INT");
							}
							printf("%c", 'A' + j2);
						}
					}
				}
				printf("\n");
			}
		}
		printf("\n");
	}
	printf("\n");
}

pci_index( int did, int vid, int busnum, int devfuncnum )
{
	unsigned	bnum, dfnum, pindex;

	for( pindex = 0; pindex < 5; pindex++ ) {
		if( _CA_PCI_Find_Device( did, vid, pindex, &bnum, &dfnum) != PCI_SUCCESS ) {
			break;
		}
		if( busnum == bnum && devfuncnum == dfnum ) {
			printf("PCI index      = %xh\n", pindex );
			break;
		}
	}
}